/* 
	*Autor: Krzysztof Gomulski
	*Grupa dziekaska: 4E2
	*Nr albumu: 214414
	*Data: 20.04.2009
	*Temat: Zastosowanie biblioteki boost::bimap

		Biblioteka boost::bimap jest rozwiniciem znanej z biblioteki standadrowej: std::map.
	Podczas gdy ta druga umoliwia tylko jednostronn relacje element - klucz, w boost::bimap ta relacja jest obustrona. 
	Przechowujemy tu w mapach (prawej i lewej) elementy, ktre jedoczenie s kluczami dla swoich kluczy, co w dalszej czci
	jest demonstrowane. Bimap jest w wikszoci zgodna z bibliotek standardow i obiekty typu std::map mona bezpiecznie
	rzutowa na lew bd praw cz obiektu boost::bimap.
*/

#include <boost/config.hpp>

#include <iostream>
#include <string>
#include <boost/bimap.hpp>
#include <boost/bimap/unordered_set_of.hpp>
#include <boost/bimap/multiset_of.hpp>
#include <boost/bimap/list_of.hpp>
#include <boost/bimap/set_of.hpp>
#include <boost/bimap/tags/tagged.hpp>
#include <boost/bimap/support/lambda.hpp>	//potrzebne przy uzyciu metod modify

struct imie {};		//struktury uzywane jako tagi
struct numer {};
struct adres {};

using namespace boost::bimaps;
using namespace boost;
using namespace std;

/*
 * Dobrym i przejrzystym przykadem wykorzystania bimapy jest implementacja prostej ksiki telefoniczej,
 * ktra umoliwia wyszukiwanie odpowiedniego numeru telefonu znajc imi osoby do ktrej on naley i na odwrt.
 * Poniszy kod umoliwia zapoznanie sie z rnymi sposobami inicjowania bimapy, ustalania relacji midzy jej elementami,
 * oraz pniejszego ich wyszukiwania
 */

int main()
{
	// Najprostszy rodzaj relacji
	{
	typedef bimap<string,long> bm_tel;
	typedef bm_tel::relation pozycja;  // typ realcji czcy obiekty;
	bm_tel telefony;
	cout<<"==== Zastosowanie biblioteki boost::bimap do stworzenia ksiki telefonicznej ==="<<endl;
	telefony.insert(pozycja("Magda",7822467));		//wypenienie bimapy
	telefony.insert(pozycja("Piotrek",785556788));
	telefony.insert(pozycja("Pawe",887342577));
	telefony.insert(pozycja("Monika",889725467));
	bm_tel::left_const_iterator i = telefony.left.find("Magda");// przeszukanie lewej poowy bimapy
	cout<<i->first<<" nr: "<<i->second<<endl;	
		/*i->first zwraca warto klucza obiektu, a i->second warto samego obiektu. 
		 * Wida to dobrze gdy bdziemy przeszukiwa praw powke bimapy, po odpowiednim klucz*/
	bm_tel::right_iterator j = telefony.right.find(889725467);
	cout<<"1. "<<j->first<<" to nr, ktry ma "<<j->second<<endl;
		//To samo otrzymujemy przy wywoaniu:
	cout<<"2. "<<889725467<<" to nr, ktry ma "<<telefony.right.at(889725467)<<endl;
		/* Podobny efekt uzyskamy tu stosujc metode project_left().Zwraca on iteretor elementu lecego po przeciwnej stronie bimapy
			ni ten do ktrego odwouje si jej argument, ale powizanego z nim relacj bimapy.Oczywicie istnieje take metoda pproject_right()*/
	bm_tel::left_iterator tel_iter = telefony.project_left(j);
	cout<<"3. "<<tel_iter->second<<" to nr, ktry ma "<<tel_iter->first<<endl;
		/*w tym przypadku bimapa szereguje malejco swoje elementy (relacja domylna), czyli moemy wypisa alfabetycznie liste osb i ich numery*/
	cout<<endl<<"- Lista alfabetyczna:"<<endl;		
	for(bm_tel::left_const_iterator ie = telefony.left.end(), ip= telefony.left.begin();ip!=ie;++ip)
		cout<<ip->first<<" nr: "<<ip->second<<endl;
	
		/*Bimapa dostarcza take moliwoci przestawiania swoich elementw, niedostpnej w std::map i niezgodnej z STL. Jednak jest to bardzo
		 * bardzo przydatna funkcja i bezpieczna. Kade wyrzucenie wyjatku (take zdefiniwanego przez programiste) przy jej wywoaniu
		 * pozostawia bimap bez zmian. Jej jedyn wad jest to e jest przy tym kosztowna i kada modyfikacja elenemtu powoduje dwukrotne 
		 * kopiowanie elementu.*/		
		cout<<endl<<"===Modyfikacje==="<<endl;	
	bm_tel::left_iterator i_paw = telefony.left.find("Pawe");  //wyszukanie elementu do zmiany
	bool zmiana = telefony.left.replace_data(i_paw, 8888888);	// zmiana numeru telefonu czyli wartoci elementu w bimapie  
	assert(zmiana); 												//sprawdzenie poprawnoci zmian
	bm_tel::right_iterator i_pio = telefony.right.find(785556788); // wyszukanie elementu z drugiej strony
	zmiana = telefony.right.replace_key(i_pio,1111111111);		//zmiana klucza
	bool zmiana2 = telefony.right.replace_data(i_pio,"Zenek, ktory byl Piotrkiem"); //zmiana watroci
	assert(zmiana & zmiana2);
		/*Szybsz ale znacznie mniej bezpiecz wersj powyszej funkcji jest metoda modify_key i modify_data.
		 * Jej gwn wad jest to e przy nieudanej modyfikacji elementu jest on czsto usuwany z bimapy*/
	bm_tel::left_iterator i_mon = telefony.left.find("Monika");  
	zmiana = telefony.left.modify_data(i_mon, _data=22222);	 // inne przekazanie argumentw ni w replace
	assert(zmiana); 							//zalecane sprawdzenie poprawnoci
	cout<<endl<<"- Lista po zmianach:"<<endl;		// wypisane nowej listy
	for(bm_tel::left_const_iterator ie = telefony.left.end(), ip= telefony.left.begin();ip!=ie;++ip)
		cout<<ip->first<<" nr: "<<ip->second<<endl;
	}
		/* W kontenerze bimap moemy podobnie jak w map z biblioteki standardowej moemy mie kontrole nad typem kolekcji
		 w niej przechowywanej. Rnica jednak polaga na tym e bimap umoliwa tak kontrole nad swoimi obiema skadowymi
		 a nie tylko nad jedn. Dziki temu przeszukiwanie bimapy jest efektywniejsze*/
	{
	cout<<endl<<"===Wasne relacje==="<<endl;	
	typedef bimap< unordered_set_of<string>,list_of<long> > bm_t; 
	typedef bm_t::value_type position;
	bm_t tel;
	tel.insert(position("Magda",7822467));
	tel.insert(position("Piotrek",785556788));
	tel.insert(position("Pawe",887342577));
	tel.insert(position("Monika",889725467));
	for(bm_t::right_const_iterator ie2 = tel.right.end(), ip2= tel.right.begin();ip2!=ie2;++ip2)
		cout<<ip2->first<<" nr: "<<ip2->second<<endl;
		/*Powyszy kod spowoduje wypisanie elementw bimapy w kolejnoci jakiej zostay do niej dodane, poniewa jej prawa strona jest
		 list. Analogiczny kod dla lewej strony bimapy wypisze jej elementy w losowej kolejnoci*/
	}
		/*Bimapa oferuje rwnie  moliwo zdefiniowania typu caej kolekcji niezalenie od typu elementw  w niej przechowywanej.
		 * Przy odpowiednim doborze tych typw, mona znaczco poprawi szybko przeszukiwania bimapy. Na przykad najszybciej
		 * robi to bimapa ktra wie elementu typu unordered_set,ale nie mona jej iterowa. ten problem rozwizuje ustawienie par elementw w list.*/
	{
	typedef bimap<
		unordered_set_of<string>,
		unordered_set_of<long>,
		list_of_relation			// typ relacji
		>bm_tel_list;
	typedef bm_tel_list::value_type pozycja_l;
	bm_tel_list tel_list;
	tel_list.push_back(pozycja_l("Magda",7822467));		// ze wzgldu na typ relacji potrzeba tu uy 'push_back' zamiast 'insert'
	tel_list.push_back(pozycja_l("Pawe",887342577));
	/*Inne typy relacji to : left_based,right_based,set_of_ralation<>,muliset_of_relation<>,unordered_set_of_relation<>,
	unordered_multiset_of_relation<>,list_of_relation,vector_of_relation,unconstrained_set_of_relation */
	}
	
	{	/*Inn przydatn funkcj jest przypisywanie do danej pary elementw odpowiedniego komentarza - tagowanie.*/
	cout<<endl<<"===Uycie tagw==="<<endl;	
	typedef bimap<
		multiset_of<tagged<string, imie> >,      // 'imie' to odpowiedni, definiowany przez programiste typ sucy do identywikacji tej skdowej bimapy 
		set_of<tagged<long,numer> >,
		with_info<tagged<string,adres> >		// okrelenie typu tagu dla pary elementw tej bimapy
	> bm_tel_tag;
	typedef bm_tel_tag::value_type position_tag;
	bm_tel_tag tel_tag;
	tel_tag.insert(position_tag("Magda",7822467,"Malinowa 15"));
	tel_tag.insert(position_tag("Pawe",887342577,"Akacjowa 23"));
	bm_tel_tag::map_by<imie>::iterator it=tel_tag.by<imie>().find("Pawe"); // dostp przez iterator 
	cout<<"Pojedyczy wpis 1 : "<<it->get<imie>()<<" nr: "<<it->get<numer>()<<" "<<it->get<adres>()<<endl;
	cout<<"Pojedyczy wpis 2 : "<<tel_tag.by<numer>().at(7822467)<<" nr: 7822467 "<<tel_tag.by<numer>().info_at(7822467)<<endl; // dostp bezporedni
	}
	
}

